import cPickle
import yaml

import numpy as np
import pandas as pd
import logging

from . import misc
from ..models import util
import urbansim.sim.simulation as sim

logger = logging.getLogger(__name__)

NETWORKS = None


def from_yaml(cfgname):
    print "Computing accessibility variables"
    cfg = yaml.load(open(misc.config(cfgname)))

    nodes = pd.DataFrame(index=NETWORKS.external_nodeids)

    node_col = cfg.get('node_col', None)

    for variable in cfg['variable_definitions']:

        name = variable["name"]
        print "Computing %s" % name

        decay = {
            "exponential": "DECAY_EXP",
            "linear": "DECAY_LINEAR",
            "flat": "DECAY_FLAT"
        }.get(variable.get("decay", "linear"))

        agg = {
            "sum": "AGG_SUM",
            "average": "AGG_AVE",
            "stddev": "AGG_STDDEV"
        }.get(variable.get("aggregation", "sum"))

        vname = variable.get("varname", None)

        radius = variable["radius"]

        dfname = variable["dataframe"]

        flds = [vname] if vname else []
        if 'add_fields' in variable:
            flds += variable['add_fields']
        if node_col:
            flds.append(node_col)
        logger.info("    Fields available to accvar =", ', '.join(flds))

        df = sim.get_table(dfname).to_frame(flds)

        if "filters" in variable:
            df = util.apply_filter_query(df, variable["filters"])
            logger.info("    Filters = %s" % variable["filters"])

        logger.info("    dataframe = %s, varname=%s" % (dfname, vname))
        logger.info("    radius = %s, aggregation = %s, decay = %s" % (
            radius, agg, decay))

        nodes[name] = NETWORKS.accvar(
            df, radius, node_ids=node_col, agg=agg, decay=decay,
            vname=vname).astype('float').values

        if "apply" in variable:
            nodes[name] = nodes[name].apply(eval(variable["apply"]))

    return nodes


class Networks:

    # flatten_nodeids is used when there is one graph to make a list of nodeids
    # rather than a list of lists - it doesn't work right now unfortunately
    def __init__(self, filenames, factors, maxdistances, twoway,
                 impedances=None, flatten_nodeids=False):
        if not filenames:
            return
        from pyaccess.pyaccess import PyAccess
        self.pya = PyAccess()
        self.pya.createGraphs(len(filenames))
        if impedances is None:
            impedances = [None] * len(filenames)
        self.nodeids = []
        self.external_nodeids = []
        for num, filename, factor, maxdistance, twoway, impedance in \
                zip(range(len(filenames)), filenames, factors, maxdistances,
                    twoway, impedances):
            net = cPickle.load(open(filename))
            if impedance is None:
                impedance = "net['edgeweights']"
            impedance = eval(impedance)
            self.pya.createGraph(
                num, net['nodeids'], net['nodes'], net['edges'],
                impedance * factor, twoway=twoway)
            if len(filenames) == 1 and flatten_nodeids:
                self.nodeids = net['nodeids']
            else:
                # these are the internal ids
                self.nodeids += zip([num] * len(net['nodeids']),
                                    range(len(net['nodeids'])))
            self.external_nodeids.append(net['nodeids'])
            self.pya.precomputeRange(maxdistance, num)

    def accvar(self, df, distance, node_ids=None, xname='x', yname='y',
               vname=None, agg="AGG_SUM", decay="DECAY_LINEAR"):
        assert self.pya  # need to generate pyaccess first
        pya = self.pya
        if isinstance(agg, str):
            agg = getattr(pya, agg)
        if isinstance(decay, str):
            decay = getattr(pya, decay)
        if vname:
            df = df.dropna(subset=[vname])
        if node_ids is None:
            xys = np.array(df[[xname, yname]], dtype="float32")
            node_ids = []
            for gno in range(pya.numgraphs):
                node_ids.append(pya.XYtoNode(xys, distance=1000, gno=gno))
        if isinstance(node_ids, str):
            l = len(df)
            df = df.dropna(subset=[node_ids])
            newl = len(df)
            if newl-l > 0:
                print "Removed %d rows because there are no node_ids" % (newl-l)
            node_ids = [df[node_ids].astype("int32").values]
        elif not isinstance(node_ids, list):
            node_ids = [node_ids]

        pya.initializeAccVars(1)
        num = 0
        aggvar = df[vname].astype('float32') if vname is not None else np.ones(
            len(df.index), dtype='float32')
        pya.initializeAccVar(num, node_ids, aggvar, preaggregate=0)
        res = []
        for gno in range(pya.numgraphs):
            res.append(pya.getAllAggregateAccessibilityVariables(
                distance, num, agg, decay, gno=gno))
        return pd.Series(
            np.concatenate(res), index=pd.MultiIndex.from_tuples(self.nodeids))

    def addnodeid(self, df):

        try:
            xys = np.array(df[['x', 'y']], dtype="float32")
        except:
            xys = np.array(df[['X', 'Y']], dtype="float32")

        for gno in range(self.pya.numgraphs):
            df['_node_id%d' % gno] = pd.Series(
                self.pya.XYtoNode(xys, gno=gno), index=df.index)
        # assign the external id as well
        df['_node_id'] = pd.Series(self.pya.getGraphIDS()[df['_node_id0'].values],
                                   index=df.index)
        return df



"""
Use the ``RelocationModel`` class to choose movers based on
relocation rates.

"""
import logging

import numpy as np
import pandas as pd

from . import util

logger = logging.getLogger(__name__)


def find_movers(choosers, rates, rate_column):
    """
    Returns an array of the indexes of the `choosers` that are slated
    to move.

    Parameters
    ----------
    choosers : pandas.DataFrame
        Table of agents from which to find movers.
    rates : pandas.DataFrame
        Table of relocation rates. Index is unused.

        Other columns describe filters on the `choosers`
        table so that different segments can have different relocation
        rates. Columns that ends with '_max' will be used to create
        a "less than" filters, columns that end with '_min' will be
        used to create "greater than or equal to" filters.
        A column with no suffix will be used to make an 'equal to' filter.

        An example `rates` structure:

        age_of_head_max  age_of_head_min
                    nan               65
                     65               40

        In this example the `choosers` table would need to have an
        'age_of_head' column on which to filter.

        nan should be used to flag filters that do not apply
        in a given row.
    rate_column : object
        Name of column in `rates` table that has relocation rates.

    Returns
    -------
    movers : pandas.Index
        Suitable for indexing `choosers` by index.

    """
    logger.debug('start: find movers for relocation')
    relocation_rates = pd.Series(
        np.zeros(len(choosers)), index=choosers.index)

    for _, row in rates.iterrows():
        indexes = util.filter_table(choosers, row, ignore={rate_column}).index
        relocation_rates.loc[indexes] = row[rate_column]

    movers = relocation_rates.index[
        relocation_rates > np.random.random(len(choosers))]
    logger.debug('picked {} movers for relocation'.format(len(movers)))
    logger.debug('finish: find movers for relocation')
    return movers


class RelocationModel(object):
    """
    Find movers within a population according to a table of
    relocation rates.

    Parameters
    ----------
    rates : pandas.DataFrame
        Table of relocation rates. Index is unused.

        Other columns describe filters on the `choosers`
        table so that different segments can have different relocation
        rates. Columns that ends with '_max' will be used to create
        a "less than" filters, columns that end with '_min' will be
        used to create "greater than or equal to" filters.
        A column with no suffix will be used to make an 'equal to' filter.

        An example `rates` structure:

        age_of_head_max  age_of_head_min
                    nan               65
                     65               40

        In this example the `choosers` table would need to have an
        'age_of_head' column on which to filter.

        nan should be used to flag filters that do not apply
        in a given row.
    rate_column : object, optional
        Name of column in `rates` table that contains relocation rates.
        If not given 'probability_of_relocating' is used.

    """
    def __init__(self, rates, rate_column=None):
        self.relocation_rates = rates
        self.rate_column = rate_column or 'probability_of_relocating'

    def find_movers(self, choosers):
        """
        Select movers from among a table of `choosers` according to the
        stored relocation rates.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table of agents from which to find movers.

        Returns
        -------
        movers : pandas.Index
            Suitable for indexing `choosers` by index.

        """
        return find_movers(choosers, self.relocation_rates, self.rate_column)



from bottle import route, response, run, hook, static_file
from urbansim.utils import yamlio
import simplejson
import numpy as np
import pandas as pd
import os
from jinja2 import Environment


@hook('after_request')
def enable_cors():
    response.headers['Access-Control-Allow-Origin'] = '*'
    response.headers['Access-Control-Allow-Headers'] = \
        'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token'

DFRAMES = {}
CONFIG = None


def get_schema():
    global DFRAMES
    return {name: list(DFRAMES[name].columns) for name in DFRAMES}


@route('/map_query/<table>/<filter>/<groupby>/<field:path>/<agg>', method="GET")
def map_query(table, filter, groupby, field, agg):
    global DFRAMES

    filter = ".query('%s')" % filter if filter != "empty" else ""

    df = DFRAMES[table]

    if field not in df.columns:
        print "Col not found, trying eval:", field
        df["eval"] = df.eval(field)
        field = "eval"

    cmd = "df%s.groupby('%s')['%s'].%s" % \
          (filter, groupby, field, agg)
    print cmd
    results = eval(cmd)
    results[results == np.inf] = np.nan
    results = yamlio.series_to_yaml_safe(results.dropna())
    results = {int(k): results[k] for k in results}
    return results


@route('/map_query/<table>/<filter>/<groupby>/<field>/<agg>', method="OPTIONS")
def ans_options(table=None, filter=None, groupby=None, field=None, agg=None):
    return {}


@route('/')
def index():
    global CONFIG
    dir = os.path.dirname(__file__)
    index = open(os.path.join(dir, 'dframe_explorer.html')).read()
    return Environment().from_string(index).render(CONFIG)


@route('/data/<filename>')
def data_static(filename):
    return static_file(filename, root='./data')


def start(views,
          center=[37.7792, -122.2191],
          zoom=11,
          shape_json='data/zones.json',
          geom_name='ZONE_ID',  # from JSON file
          join_name='zone_id',  # from data frames
          precision=8,
          port=8765,
          host='localhost',
          testing=False):
    """
    Start the web service which serves the Pandas queries and generates the
    HTML for the map.  You will need to open a web browser and navigate to
    http://localhost:8765 (or the specified port)

    Parameters
    ----------
    views : Python dictionary
        This is the data that will be displayed in the maps.  Keys are strings
        (table names) and values are dataframes.  Each data frame should have a
        column with the name specified as join_name below
    center : a Python list with two floats
        The initial latitude and longitude of the center of the map
    zoom : int
        The initial zoom level of the map
    shape_json : str
        The path to the geojson file which contains that shapes that will be
        displayed
    geom_name : str
        The field name from the JSON file which contains the id of the geometry
    join_name : str
        The column name from the dataframes passed as views (must be in each
        view) which joins to geom_name in the shapes
    precision : int
        The precision of values to show in the legend on the map
    port : int
        The port for the web service to respond on
    host : str
        The hostname to run the web service from
    testing : bool
        Whether to print extra debug information

    Returns
    -------
    Does not return - takes over control of the thread and responds to
    queries from a web browser
    """

    global DFRAMES, CONFIG
    DFRAMES = {str(k): views[k] for k in views}

    config = {
        'center': str(center),
        'zoom': zoom,
        'shape_json': shape_json,
        'geom_name': geom_name,
        'join_name': join_name,
        'precision': precision
    }

    for k in views:
        if join_name not in views[k].columns:
            raise Exception("Join name must be present on all dataframes - "
                            "'%s' not present on '%s'" % (join_name, k))

    config['schema'] = simplejson.dumps(get_schema())

    CONFIG = config

    if testing:
        return

    run(host=host, port=port, debug=True)



import pandas as pd
import numpy as np


class Developer(object):
    """
    Pass the dataframe that is returned by feasibility here

    Can also be a dictionary where keys are building forms and values are
    the individual data frames returned by the proforma lookup routine.

    """

    def __init__(self, feasibility):
        if isinstance(feasibility, dict):
            feasibility = pd.concat(feasibility.values(), keys=feasibility.keys(), axis=1)
        self.feasibility = feasibility

    @staticmethod
    def _max_form(f, colname):
        """
        Assumes dataframe with hierarchical columns with first index equal to the
        use and second index equal to the attribute.

        e.g. f.columns equal to::

            mixedoffice   building_cost
                          building_revenue
                          building_size
                          max_profit
                          max_profit_far
                          total_cost
            industrial    building_cost
                          building_revenue
                          building_size
                          max_profit
                          max_profit_far
                          total_cost
        """
        df = f.stack(level=0)[[colname]].stack().unstack(level=1).reset_index(level=1, drop=True)
        return df.idxmax(axis=1)

    def keep_form_with_max_profit(self, forms=None):
        """
        This converts the dataframe, which shows all profitable forms,
        to the form with the greatest profit, so that more profitable
        forms outcompete less profitable forms.

        Parameters
        ----------
        forms: list of strings
            List of forms which compete which other.  Can leave some out.

        Returns
        -------
        Nothing.  Goes from a multi-index to a single index with only the
        most profitable form.
        """
        f = self.feasibility

        if forms is not None:
            f = f[forms]

        mu = self._max_form(f, "max_profit")
        indexes = [tuple(x) for x in mu.reset_index().values]
        df = f.stack(level=0).loc[indexes]
        df.index.names = ["parcel_id", "form"]
        df = df.reset_index(level=1)
        return df

    @staticmethod
    def compute_units_to_build(num_agents, num_units, target_vacancy):
        """
        Compute number of units to build to match target vacancy.

        Parameters
        ----------
        num_agents : int
            number of agents that need units in the region
        num_units : int
            number of units in buildings
        target_vacancy : float (0-1.0)
            target vacancy rate

        Returns
        -------
        number_of_units : int
            the number of units that need to be built
        """
        print "Number of agents: {:,}".format(num_agents)
        print "Number of agent spaces: {:,}".format(int(num_units))
        assert target_vacancy < 1.0
        target_units = int(max(num_agents / (1 - target_vacancy) - num_units, 0))
        print "Current vacancy = {:.2f}".format(1 - num_agents / float(num_units))
        print "Target vacancy = {:.2f}, target of new units = {:,}".\
            format(target_vacancy, target_units)
        return target_units

    def pick(self, form, target_units, parcel_size, ave_unit_size,
             current_units, max_parcel_size=200000, min_unit_size=400,
             drop_after_build=True, residential=True, bldg_sqft_per_job=400.0):
        """
        Choose the buildings from the list that are feasible to build in
        order to match the specified demand.

        Parameters
        ----------
        form : string or list
            One or more of the building forms from the pro forma specification -
            e.g. "residential" or "mixedresidential" - these are configuration
            parameters passed previously to the pro forma.  If more than one form
            is passed the forms compete with each other (based on profitability)
            for which one gets built in order to meet demand.
        target_units : int
            The number of units to build.  For non-residential buildings this
            should be passed as the number of job spaces that need to be created.
        parcel_size : series
            The size of the parcels.  This was passed to feasibility as well,
            but should be passed here as well.  Index should be parcel_ids.
        ave_unit_size : series
            The average residential unit size around each parcel - this is
            indexed by parcel, but is usually a disaggregated version of a
            zonal or accessibility aggregation.
        bldg_sqft_per_job : float (default 400.0)
            The average square feet per job for this building form.
        min_unit_size : float
            Values less than this number in ave_unit_size will be set to this
            number.  Deals with cases where units are currently not built.
        current_units : series
            The current number of units on the parcel.  Is used to compute the
            net number of units produced by the developer model.  Many times
            the developer model is redeveloping units (demolishing them) and
            is trying to meet a total number of net units produced.
        max_parcel_size : float
            Parcels larger than this size will not be considered for
            development - usually large parcels should be specified manually
            in a development projects table.
        drop_after_build : bool
            Whether or not to drop parcels from consideration after they
            have been chosen for development.  Usually this is true so as
            to not develop the same parcel twice.
        residential: bool
            If creating non-residential buildings set this to false and
            developer will fill in job_spaces rather than residential_units

        Returns
        -------
        new_buildings : dataframe
            DataFrame of buildings to add.  These buildings are rows from the
            DataFrame that is returned from feasibility.
        """

        if isinstance(form, list):
            df = self.keep_form_with_max_profit(form)
        else:
            df = self.feasibility[form]

        # feasible buildings only for this building type
        df = df[df.max_profit_far > 0]
        ave_unit_size[ave_unit_size < min_unit_size] = min_unit_size
        df["ave_unit_size"] = ave_unit_size
        df["parcel_size"] = parcel_size
        df['current_units'] = current_units
        df = df[df.parcel_size < max_parcel_size]

        df['residential_units'] = np.round(df.residential_sqft / df.ave_unit_size)
        df['job_spaces'] = np.round(df.non_residential_sqft / bldg_sqft_per_job)

        if residential:
            df['net_units'] = df.residential_units - df.current_units
        else:
            df['net_units'] = df.job_spaces - df.current_units
        df = df[df.net_units > 0]

        if len(df) == 0:
            print "WARNING THERE ARE NO FEASIBLE BUILDING TO CHOOSE FROM"
            return

        # print "Describe of net units\n", df.net_units.describe()
        print "Sum of net units that are profitable: {:,}".\
            format(int(df.net_units.sum()))
        if df.net_units.sum() < target_units:
            print "WARNING THERE WERE NOT ENOUGH PROFITABLE UNITS TO " \
                  "MATCH DEMAND"

        choices = np.random.choice(df.index.values, size=len(df.index),
                                   replace=False,
                                   p=(df.max_profit.values / df.max_profit.sum()))
        net_units = df.net_units.loc[choices]
        tot_units = net_units.values.cumsum()
        ind = int(np.searchsorted(tot_units, target_units, side="left"))
        if target_units != 0:
            ind += 1
        ind = min(ind, len(choices))
        build_idx = choices[:ind]

        if drop_after_build:
            self.feasibility = self.feasibility.drop(build_idx)

        new_df = df.loc[build_idx]
        new_df.index.name = "parcel_id"
        return new_df.reset_index()

    @staticmethod
    def merge(old_df, new_df):
        """
        Merge two dataframes of buildings.  The old dataframe is
        usually the buildings dataset and the new dataframe is a modified
        (by the user) version of what is returned by the pick method.

        Parameters
        ----------
        old_df : dataframe
            Current set of buildings
        new_df : dataframe
            New buildings to add, usually comes from this module

        Returns
        -------
        df : dataframe
            Combined DataFrame of buildings, makes sure indexes don't overlap
        """
        maxind = np.max(old_df.index.values)
        new_df = new_df.reset_index(drop=True)
        new_df.index = new_df.index + maxind + 1
        concat_df = pd.concat([old_df, new_df], verify_integrity=True)
        concat_df.index.name = 'building_id'
        return concat_df



"""
Utilities used within the ``urbansim.models`` package.

"""
import collections
import logging
import numbers
from StringIO import StringIO
from tokenize import generate_tokens, NAME

import numpy as np
import pandas as pd
import patsy
import toolz

from ..utils.logutil import log_start_finish

logger = logging.getLogger(__name__)


def apply_filter_query(df, filters=None):
    """
    Use the DataFrame.query method to filter a table down to the
    desired rows.

    Parameters
    ----------
    df : pandas.DataFrame
    filters : list of str or str, optional
        List of filters to apply. Will be joined together with
        ' and ' and passed to DataFrame.query. A string will be passed
        straight to DataFrame.query.
        If not supplied no filtering will be done.

    Returns
    -------
    filtered_df : pandas.DataFrame

    """
    with log_start_finish('apply filter query: {!r}'.format(filters), logger):
        if filters:
            if isinstance(filters, str):
                query = filters
            else:
                query = ' and '.join(filters)
            return df.query(query)
        else:
            return df


def _filterize(name, value):
    """
    Turn a `name` and `value` into a string expression compatible
    the ``DataFrame.query`` method.

    Parameters
    ----------
    name : str
        Should be the name of a column in the table to which the
        filter will be applied.

        A suffix of '_max' will result in a "less than" filter,
        a suffix of '_min' will result in a "greater than or equal to" filter,
        and no recognized suffix will result in an "equal to" filter.
    value : any
        Value side of filter for comparison to column values.

    Returns
    -------
    filter_exp : str

    """
    if name.endswith('_min'):
        name = name[:-4]
        comp = '>='
    elif name.endswith('_max'):
        name = name[:-4]
        comp = '<'
    else:
        comp = '=='

    result = '{} {} {!r}'.format(name, comp, value)
    logger.debug(
        'converted name={} and value={} to filter {}'.format(
            name, value, result))
    return result


def filter_table(table, filter_series, ignore=None):
    """
    Filter a table based on a set of restrictions given in
    Series of column name / filter parameter pairs. The column
    names can have suffixes `_min` and `_max` to indicate
    "less than" and "greater than" constraints.

    Parameters
    ----------
    table : pandas.DataFrame
        Table to filter.
    filter_series : pandas.Series
        Series of column name / value pairs of filter constraints.
        Columns that ends with '_max' will be used to create
        a "less than" filters, columns that end with '_min' will be
        used to create "greater than or equal to" filters.
        A column with no suffix will be used to make an 'equal to' filter.
    ignore : sequence of str, optional
        List of column names that should not be used for filtering.

    Returns
    -------
    filtered : pandas.DataFrame

    """
    with log_start_finish('filter table', logger):
        ignore = ignore if ignore else set()

        filters = [_filterize(name, val)
                   for name, val in filter_series.iteritems()
                   if not (name in ignore or
                           (isinstance(val, numbers.Number) and
                            np.isnan(val)))]

        return apply_filter_query(table, filters)


def concat_indexes(indexes):
    """
    Concatenate a sequence of pandas Indexes.

    Parameters
    ----------
    indexes : sequence of pandas.Index

    Returns
    -------
    pandas.Index

    """
    return pd.Index(np.concatenate(indexes))


def has_constant_expr(expr):
    """
    Report whether a model expression has constant specific term.
    That is, a term explicitly specying whether the model should or
    should not include a constant. (e.g. '+ 1' or '- 1'.)

    Parameters
    ----------
    expr : str
        Model expression to check.

    Returns
    -------
    has_constant : bool

    """
    def has_constant(node):
        if node.type == 'ONE':
            return True

        for n in node.args:
            if has_constant(n):
                return True

        return False

    return has_constant(patsy.parse_formula.parse_formula(expr))


def str_model_expression(expr, add_constant=True):
    """
    We support specifying model expressions as strings, lists, or dicts;
    but for use with patsy and statsmodels we need a string.
    This function will take any of those as input and return a string.

    Parameters
    ----------
    expr : str, iterable, or dict
        A string will be returned unmodified except to add or remove
        a constant.
        An iterable sequence will be joined together with ' + '.
        A dictionary should have ``right_side`` and, optionally,
        ``left_side`` keys. The ``right_side`` can be a list or a string
        and will be handled as above. If ``left_side`` is present it will
        be joined with ``right_side`` with ' ~ '.
    add_constant : bool, optional
        Whether to add a ' + 1' (if True) or ' - 1' (if False) to the model.
        If the expression already has a '+ 1' or '- 1' this option will be
        ignored.

    Returns
    -------
    model_expression : str
        A string model expression suitable for use with statsmodels and patsy.

    """
    if not isinstance(expr, str):
        if isinstance(expr, collections.Mapping):
            left_side = expr.get('left_side')
            right_side = str_model_expression(expr['right_side'], add_constant)
        else:
            # some kind of iterable like a list
            left_side = None
            right_side = ' + '.join(expr)

        if left_side:
            model_expression = ' ~ '.join((left_side, right_side))
        else:
            model_expression = right_side

    else:
        model_expression = expr

    if not has_constant_expr(model_expression):
        if add_constant:
            model_expression += ' + 1'
        else:
            model_expression += ' - 1'

    logger.debug(
        'converted expression: {!r} to model: {!r}'.format(
            expr, model_expression))
    return model_expression


def sorted_groupby(df, groupby):
    """
    Perform a groupby on a DataFrame using a specific column
    and assuming that that column is sorted.

    Parameters
    ----------
    df : pandas.DataFrame
    groupby : object
        Column name on which to groupby. This column must be sorted.

    Returns
    -------
    generator
        Yields pairs of group_name, DataFrame.

    """
    start = 0
    prev = df[groupby].iloc[start]
    for i, x in enumerate(df[groupby]):
        if x != prev:
            yield prev, df.iloc[start:i]
            prev = x
            start = i
    # need to send back the last group
    yield prev, df.iloc[start:]


def columns_in_filters(filters):
    """
    Returns a list of the columns used in a set of query filters.

    Parameters
    ----------
    filters : list of str or str
        List of the filters as passed passed to ``apply_filter_query``.

    Returns
    -------
    columns : list of str
        List of all the strings mentioned in the filters.

    """
    if not filters:
        return []

    if not isinstance(filters, str):
        filters = ' '.join(filters)

    columns = []
    reserved = {'and', 'or', 'in', 'not'}

    for toknum, tokval, _, _, _ in generate_tokens(StringIO(filters).readline):
        if toknum == NAME and tokval not in reserved:
            columns.append(tokval)

    return list(toolz.unique(columns))


def _tokens_from_patsy(node):
    """
    Yields all the individual tokens from within a patsy formula
    as parsed by patsy.parse_formula.parse_formula.

    Parameters
    ----------
    node : patsy.parse_formula.ParseNode

    """
    for n in node.args:
        for t in _tokens_from_patsy(n):
            yield t

    if node.token:
        yield node.token


def columns_in_formula(formula):
    """
    Returns the names of all the columns used in a patsy formula.

    Parameters
    ----------
    formula : str, iterable, or dict
        Any formula construction supported by ``str_model_expression``.

    Returns
    -------
    columns : list of str

    """
    if formula is None:
        return []

    formula = str_model_expression(formula, add_constant=False)
    columns = []

    tokens = map(
        lambda x: x.extra,
        toolz.remove(
            lambda x: x.extra is None,
            _tokens_from_patsy(patsy.parse_formula.parse_formula(formula))))

    for tok in tokens:
        # if there are parentheses in the expression we
        # want to drop them and everything outside
        # and start again from the top
        if '(' in tok:
            start = tok.find('(') + 1
            fin = tok.rfind(')')
            columns.extend(columns_in_formula(tok[start:fin]))
        else:
            for toknum, tokval, _, _, _ in generate_tokens(
                    StringIO(tok).readline):
                if toknum == NAME:
                    columns.append(tokval)

    return list(toolz.unique(columns))



import os.path

from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages

# read README as the long description
if os.path.exists('README'):
    with open('README', 'r') as f:
        long_description = f.read()
else:
    long_description = None

setup(
    name='urbansim',
    version='1.0',
    description='Tool for modeling metropolitan real estate markets',
    long_description=long_description,
    author='Synthicity',
    author_email='ffoti@berkeley.edu',
    license='BSD',
    url='https://github.com/synthicity/urbansim',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: BSD License'
    ],
    package_data={
        '': ['*.html'],
    },
    packages=find_packages(exclude=['*.tests']),
    install_requires=[
        'bottle>=0.12.5',
        'matplotlib>=1.3.1',
        'numpy>=1.8.0',
        'pandas>=0.13.1',
        'patsy>=0.2.1',
        'prettytable>=0.7.2',
        'pyyaml>=3.10',
        'scipy>=0.13.3',
        'simplejson>=3.3.3',
        'statsmodels>=0.5.0',
        'tables>=3.1.0',
        'toolz>=0.7.0'
    ]
)



"""
Use the ``MNLLocationChoiceModel`` class to train a choice module using
multinomial logit and make subsequent choice predictions.

"""
from __future__ import print_function, division

import logging

import numpy as np
import pandas as pd
from patsy import dmatrix
from prettytable import PrettyTable
import toolz

from . import util
from ..exceptions import ModelEvaluationError
from ..urbanchoice import interaction, mnl
from ..utils import yamlio
from ..utils.logutil import log_start_finish

logger = logging.getLogger(__name__)


def unit_choice(chooser_ids, alternative_ids, probabilities):
    """
    Have a set of choosers choose from among alternatives according
    to a probability distribution. Choice is binary: each
    alternative can only be chosen once.

    Parameters
    ----------
    chooser_ids : 1d array_like
        Array of IDs of the agents that are making choices.
    alternative_ids : 1d array_like
        Array of IDs of alternatives among which agents are making choices.
    probabilities : 1d array_like
        The probability that an agent will choose an alternative.
        Must be the same shape as `alternative_ids`. Unavailable
        alternatives should have a probability of 0.

    Returns
    -------
    choices : pandas.Series
        Mapping of chooser ID to alternative ID. Some choosers
        will map to a nan value when there are not enough alternatives
        for all the choosers.

    """
    chooser_ids = np.asanyarray(chooser_ids)
    alternative_ids = np.asanyarray(alternative_ids)
    probabilities = np.asanyarray(probabilities)

    logger.debug(
        'start: unit choice with {} choosers and {} alternatives'.format(
            len(chooser_ids), len(alternative_ids)))

    choices = pd.Series([np.nan] * len(chooser_ids), index=chooser_ids)

    if probabilities.sum() == 0:
        # return all nan if there are no available units
        return choices

    # probabilities need to sum to 1 for np.random.choice
    probabilities = probabilities / probabilities.sum()

    # need to see if there are as many available alternatives as choosers
    n_available = np.count_nonzero(probabilities)
    n_choosers = len(chooser_ids)
    n_to_choose = n_choosers if n_choosers < n_available else n_available

    chosen = np.random.choice(
        alternative_ids, size=n_to_choose, replace=False, p=probabilities)

    # if there are fewer available units than choosers we need to pick
    # which choosers get a unit
    if n_to_choose == n_available:
        chooser_ids = np.random.choice(
            chooser_ids, size=n_to_choose, replace=False)

    choices[chooser_ids] = chosen

    logger.debug('finish: unit choice')
    return choices


class MNLLocationChoiceModel(object):
    """
    A location choice model with the ability to store an estimated
    model and predict new data based on the model.
    Based on multinomial logit.

    Parameters
    ----------
    model_expression : str, iterable, or dict
        A patsy model expression. Should contain only a right-hand side.
    sample_size : int
        Number of choices to sample for estimating the model.
    choosers_fit_filters : list of str, optional
        Filters applied to choosers table before fitting the model.
    choosers_predict_filters : list of str, optional
        Filters applied to the choosers table before calculating
        new data points.
    alts_fit_filters : list of str, optional
        Filters applied to the alternatives table before fitting the model.
    alts_predict_filters : list of str, optional
        Filters applied to the alternatives table before calculating
        new data points.
    interaction_predict_filters : list of str, optional
        Filters applied to the merged choosers/alternatives table
        before predicting agent choices.
    estimation_sample_size : int, optional, whether to sample choosers
        during estimation (needs to be applied after choosers_fit_filters)
    choice_column : optional
        Name of the column in the `alternatives` table that choosers
        should choose. e.g. the 'building_id' column. If not provided
        the alternatives index is used.
    name : optional
        Optional descriptive name for this model that may be used
        in output.

    """
    def __init__(self, model_expression, sample_size,
                 choosers_fit_filters=None, choosers_predict_filters=None,
                 alts_fit_filters=None, alts_predict_filters=None,
                 interaction_predict_filters=None,
                 estimation_sample_size=None,
                 choice_column=None, name=None):
        self.model_expression = model_expression
        self.sample_size = sample_size
        self.choosers_fit_filters = choosers_fit_filters
        self.choosers_predict_filters = choosers_predict_filters
        self.alts_fit_filters = alts_fit_filters
        self.alts_predict_filters = alts_predict_filters
        self.interaction_predict_filters = interaction_predict_filters
        self.estimation_sample_size = estimation_sample_size
        self.choice_column = choice_column
        self.name = name if name is not None else 'MNLLocationChoiceModel'
        self.sim_pdf = None

        self.log_likelihoods = None
        self.fit_parameters = None

    @classmethod
    def from_yaml(cls, yaml_str=None, str_or_buffer=None):
        """
        Create a LocationChoiceModel instance from a saved YAML configuration.
        Arguments are mutally exclusive.

        Parameters
        ----------
        yaml_str : str, optional
            A YAML string from which to load model.
        str_or_buffer : str or file like, optional
            File name or buffer from which to load YAML.

        Returns
        -------
        MNLLocationChoiceModel

        """
        cfg = yamlio.yaml_to_dict(yaml_str, str_or_buffer)

        model = cls(
            cfg['model_expression'],
            cfg['sample_size'],
            choosers_fit_filters=cfg.get('choosers_fit_filters', None),
            choosers_predict_filters=cfg.get('choosers_predict_filters', None),
            alts_fit_filters=cfg.get('alts_fit_filters', None),
            alts_predict_filters=cfg.get('alts_predict_filters', None),
            interaction_predict_filters=cfg.get(
                'interaction_predict_filters', None),
            estimation_sample_size=cfg.get('estimation_sample_size', None),
            choice_column=cfg.get('choice_column', None),
            name=cfg.get('name', None)
        )

        if cfg.get('log_likelihoods', None):
            model.log_likelihoods = cfg['log_likelihoods']
        if cfg.get('fit_parameters', None):
            model.fit_parameters = pd.DataFrame(cfg['fit_parameters'])

        logger.debug('loaded LCM model {} from YAML'.format(model.name))
        return model

    @property
    def str_model_expression(self):
        """
        Model expression as a string suitable for use with patsy/statsmodels.

        """
        return util.str_model_expression(
            self.model_expression, add_constant=False)

    def fit(self, choosers, alternatives, current_choice):
        """
        Fit and save model parameters based on given data.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing,
            e.g. buildings.
        current_choice : pandas.Series or any
            A Series describing the `alternatives` currently chosen
            by the `choosers`. Should have an index matching `choosers`
            and values matching the index of `alternatives`.

            If a non-Series is given it should be a column in `choosers`.

        Returns
        -------
        log_likelihoods : dict
            Dict of log-liklihood values describing the quality of the
            model fit. Will have keys 'null', 'convergence', and 'ratio'.

        """
        logger.debug('start: fit LCM model {}'.format(self.name))

        if not isinstance(current_choice, pd.Series):
            current_choice = choosers[current_choice]

        choosers = util.apply_filter_query(choosers, self.choosers_fit_filters)

        if self.estimation_sample_size:
            choosers = choosers.loc[np.random.choice(
                choosers.index,
                min(self.estimation_sample_size, len(choosers)),
                replace=False)]

        current_choice = current_choice.loc[choosers.index]

        alternatives = util.apply_filter_query(
            alternatives, self.alts_fit_filters)

        _, merged, chosen = interaction.mnl_interaction_dataset(
            choosers, alternatives, self.sample_size, current_choice)
        model_design = dmatrix(
            self.str_model_expression, data=merged, return_type='dataframe')

        if len(merged) != model_design.as_matrix().shape[0]:
            raise ModelEvaluationError(
                'Estimated data does not have the same length as input.  '
                'This suggests there are null values in one or more of '
                'the input columns.')

        self.log_likelihoods, self.fit_parameters = mnl.mnl_estimate(
            model_design.as_matrix(), chosen, self.sample_size)
        self.fit_parameters.index = model_design.columns

        logger.debug('finish: fit LCM model {}'.format(self.name))
        return self.log_likelihoods

    @property
    def fitted(self):
        """
        True if model is ready for prediction.

        """
        return self.fit_parameters is not None

    def assert_fitted(self):
        """
        Raises `RuntimeError` if the model is not ready for prediction.

        """
        if not self.fitted:
            raise RuntimeError('Model has not been fit.')

    def report_fit(self):
        """
        Print a report of the fit results.

        """
        if not self.fitted:
            print('Model not yet fit.')
            return

        print('Null Log-liklihood: {0:.3f}'.format(
            self.log_likelihoods['null']))
        print('Log-liklihood at convergence: {0:.3f}'.format(
            self.log_likelihoods['convergence']))
        print('Log-liklihood Ratio: {0:.3f}\n'.format(
            self.log_likelihoods['ratio']))

        tbl = PrettyTable(
            ['Component', ])
        tbl = PrettyTable()

        tbl.add_column('Component', self.fit_parameters.index.values)
        for col in ('Coefficient', 'Std. Error', 'T-Score'):
            tbl.add_column(col, self.fit_parameters[col].values)

        tbl.align['Component'] = 'l'
        tbl.float_format = '.3'

        print(tbl)

    def predict(self, choosers, alternatives, debug=False):
        """
        Choose from among alternatives for a group of agents.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
            Only the first item in this table is used for determining
            agent probabilities of choosing alternatives.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing.
        debug : bool
            If debug is set to true, will set the variable "sim_pdf" on
            the object to store the probabilities for mapping of the
            outcome.

        Returns
        -------
        choices : pandas.Series
            Mapping of chooser ID to alternative ID. Some choosers
            will map to a nan value when there are not enough alternatives
            for all the choosers.

        """
        self.assert_fitted()
        logger.debug('start: predict LCM model {}'.format(self.name))

        choosers = util.apply_filter_query(
            choosers, self.choosers_predict_filters)
        alternatives = util.apply_filter_query(
            alternatives, self.alts_predict_filters)

        if len(choosers) == 0:
            return pd.Series()

        if len(alternatives) == 0:
            return pd.Series(index=choosers.index)

        # TODO: only using 1st item in choosers for determining probabilities.
        # Need to expand options around this.
        num_choosers = 1
        _, merged, _ = interaction.mnl_interaction_dataset(
            choosers.head(num_choosers), alternatives, len(alternatives))
        merged = util.apply_filter_query(
            merged, self.interaction_predict_filters)
        model_design = dmatrix(
            self.str_model_expression, data=merged, return_type='dataframe')

        if len(merged) != model_design.as_matrix().shape[0]:
            raise ModelEvaluationError(
                'Simulated data does not have the same length as input.  '
                'This suggests there are null values in one or more of '
                'the input columns.')

        coeffs = [self.fit_parameters['Coefficient'][x]
                  for x in model_design.columns]

        # probabilities are returned from mnl_simulate as a 2d array
        # and need to be flatted for use in unit_choice.
        probabilities = mnl.mnl_simulate(
            model_design.as_matrix(),
            coeffs,
            numalts=len(merged), returnprobs=True).flatten()

        if debug:
            # when we're not doing 1st item of choosers, this will break!
            assert num_choosers == 1
            self.sim_pdf = pd.Series(probabilities, index=merged.index)

        # figure out exactly which things from which choices are drawn
        alt_choices = (
            merged[self.choice_column] if self.choice_column else merged.index)

        result = unit_choice(
            choosers.index, alt_choices, probabilities)
        logger.debug('finish: predict LCM model {}'.format(self.name))
        return result

    def to_dict(self):
        """
        Return a dict respresentation of an MNLLocationChoiceModel
        instance.

        """
        return {
            'model_type': 'locationchoice',
            'model_expression': self.model_expression,
            'sample_size': self.sample_size,
            'name': self.name,
            'choosers_fit_filters': self.choosers_fit_filters,
            'choosers_predict_filters': self.choosers_predict_filters,
            'alts_fit_filters': self.alts_fit_filters,
            'alts_predict_filters': self.alts_predict_filters,
            'interaction_predict_filters': self.interaction_predict_filters,
            'estimation_sample_size': self.estimation_sample_size,
            'choice_column': self.choice_column,
            'fitted': self.fitted,
            'log_likelihoods': self.log_likelihoods,
            'fit_parameters': (yamlio.frame_to_yaml_safe(self.fit_parameters)
                               if self.fitted else None)
        }

    def to_yaml(self, str_or_buffer=None):
        """
        Save a model respresentation to YAML.

        Parameters
        ----------
        str_or_buffer : str or file like, optional
            By default a YAML string is returned. If a string is
            given here the YAML will be written to that file.
            If an object with a ``.write`` method is given the
            YAML will be written to that object.

        Returns
        -------
        j : str
            YAML is string if `str_or_buffer` is not given.

        """
        logger.debug('serializing LCM model {} to YAML'.format(self.name))
        return yamlio.convert_to_yaml(self.to_dict(), str_or_buffer)

    def choosers_columns_used(self):
        """
        Columns from the choosers table that are used for filtering.

        """
        return list(toolz.unique(toolz.concatv(
            util.columns_in_filters(self.choosers_predict_filters),
            util.columns_in_filters(self.choosers_fit_filters))))

    def alts_columns_used(self):
        """
        Columns from the alternatives table that are used for filtering.

        """
        return list(toolz.unique(toolz.concatv(
            util.columns_in_filters(self.alts_predict_filters),
            util.columns_in_filters(self.alts_fit_filters))))

    def interaction_columns_used(self):
        """
        Columns from the interaction dataset used for filtering and in
        the model. These may come originally from either the choosers or
        alternatives tables.

        """
        return list(toolz.unique(toolz.concatv(
            util.columns_in_filters(self.interaction_predict_filters),
            util.columns_in_formula(self.model_expression))))

    def columns_used(self):
        """
        Columns from any table used in the model. May come from either
        the choosers or alternatives tables.

        """
        return list(toolz.unique(toolz.concatv(
            self.choosers_columns_used(),
            self.alts_columns_used(),
            self.interaction_columns_used())))

    @classmethod
    def fit_from_cfg(cls, choosers, chosen_fname, alternatives, cfgname):
        """
        Parameters
        ----------
        choosers : DataFrame
            A dataframe of rows of agents which have locations assigned.
        chosen_fname : string
            A string indicating the column in the choosers dataframe which
            gives which location the choosers have chosen.
        alternatives : DataFrame
            A dataframe of locations which should include the chosen locations
            from the choosers dataframe as well as some other locations from
            which to sample.  Values in choosers[chosen_fname] should index
            into the alternatives dataframe.
        cfgname : string
            The name of the yaml config file from which to read the location
            choice model.

        Returns
        -------
        lcm : MNLLocationChoiceModel which was used to fit
        """
        logger.debug('start: fit from configuration {}'.format(cfgname))
        lcm = cls.from_yaml(str_or_buffer=cfgname)
        lcm.fit(choosers, alternatives, choosers[chosen_fname])
        lcm.report_fit()
        lcm.to_yaml(str_or_buffer=cfgname)
        logger.debug('finish: fit from configuration {}'.format(cfgname))
        return lcm

    @classmethod
    def predict_from_cfg(cls, movers, locations, cfgname,
                         location_ratio=2.0, debug=False):
        """
        Simulate the location choices for the specified choosers

        Parameters
        ----------
        movers : DataFrame
            A dataframe of agents doing the choosing.
        locations : DataFrame
            A dataframe of locations which the choosers are locating in and which
            have a supply.
        cfgname : string
            The name of the yaml config file from which to read the location
            choice model.
        location_ratio : float
            Above the location ratio (default of 2.0) of locations to choosers, the
            locations will be sampled to meet this ratio (for performance reasons).
        debug : boolean, optional (default False)
            Whether to generate debug information on the model.

        Returns
        -------
        choices : pandas.Series
            Mapping of chooser ID to alternative ID. Some choosers
            will map to a nan value when there are not enough alternatives
            for all the choosers.
        lcm : MNLLocationChoiceModel which was used to predict
        """
        logger.debug('start: predict from configuration {}'.format(cfgname))
        lcm = cls.from_yaml(str_or_buffer=cfgname)

        if len(locations) > len(movers) * location_ratio:
            logger.info("Location ratio exceeded: %d locations and only %d choosers" %
                        (len(locations), len(movers)))
            idxes = np.random.choice(locations.index, size=len(movers) * location_ratio,
                                     replace=False)
            locations = locations.loc[idxes]
            logger.info("  after sampling %d locations are available\n" % len(locations))

        new_units = lcm.predict(movers, locations, debug=debug)
        print("Assigned %d choosers to new units" % len(new_units.dropna()))
        logger.debug('finish: predict from configuration {}'.format(cfgname))
        return new_units, lcm


class MNLLocationChoiceModelGroup(object):
    """
    Manages a group of location choice models that refer to different
    segments of choosers.

    Model names must match the segment names after doing a pandas groupby.

    Parameters
    ----------
    segmentation_col
        Name of a column in the table of choosers. Will be used to perform
        a pandas groupby on the choosers table.

    """
    def __init__(self, segmentation_col, name=None):
        self.segmentation_col = segmentation_col
        self.name = name if name is not None else 'MNLLocationChoiceModelGroup'
        self.models = {}

    def add_model(self, model):
        """
        Add an MNLLocationChoiceModel instance.

        Parameters
        ----------
        model : MNLLocationChoiceModel
            Should have a ``.name`` attribute matching one of the segments
            in the choosers table.

        """
        logger.debug(
            'adding model {} to LCM group {}'.format(model.name, self.name))
        self.models[model.name] = model

    def add_model_from_params(
            self, name, model_expression, sample_size,
            choosers_fit_filters=None, choosers_predict_filters=None,
            alts_fit_filters=None, alts_predict_filters=None,
            interaction_predict_filters=None, estimation_sample_size=None,
            choice_column=None):
        """
        Add a model by passing parameters through to MNLLocationChoiceModel.

        Parameters
        ----------
        name
            Must match a segment in the choosers table.
        model_expression : str, iterable, or dict
            A patsy model expression. Should contain only a right-hand side.
        sample_size : int
            Number of choices to sample for estimating the model.
        choosers_fit_filters : list of str, optional
            Filters applied to choosers table before fitting the model.
        choosers_predict_filters : list of str, optional
            Filters applied to the choosers table before calculating
            new data points.
        alts_fit_filters : list of str, optional
            Filters applied to the alternatives table before fitting the model.
        alts_predict_filters : list of str, optional
            Filters applied to the alternatives table before calculating
            new data points.
        interaction_predict_filters : list of str, optional
            Filters applied to the merged choosers/alternatives table
            before predicting agent choices.
        estimation_sample_size : int, optional, whether to sample choosers
            during estimation (needs to be applied after choosers_fit_filters)
        choice_column : optional
            Name of the column in the `alternatives` table that choosers
            should choose. e.g. the 'building_id' column. If not provided
            the alternatives index is used.

        """
        logger.debug('adding model {} to LCM group {}'.format(name, self.name))
        self.models[name] = MNLLocationChoiceModel(
            model_expression, sample_size,
            choosers_fit_filters, choosers_predict_filters,
            alts_fit_filters, alts_predict_filters,
            interaction_predict_filters, estimation_sample_size,
            choice_column, name)

    def _iter_groups(self, data):
        """
        Iterate over the groups in `data` after grouping by
        `segmentation_col`. Skips any groups for which there
        is no model stored.

        Yields tuples of (name, df) where name is the group key
        and df is the group DataFrame.

        Parameters
        ----------
        data : pandas.DataFrame
            Must have a column with the same name as `segmentation_col`.

        """
        groups = data.groupby(self.segmentation_col)

        for name, group in groups:
            logger.debug(
                'returning group {} in LCM group {}'.format(name, self.name))
            yield name, group

    def fit(self, choosers, alternatives, current_choice):
        """
        Fit and save models based on given data after segmenting
        the `choosers` table.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
            Must have a column with the same name as the .segmentation_col
            attribute.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing,
            e.g. buildings.
        current_choice
            Name of column in `choosers` that indicates which alternative
            they have currently chosen.

        Returns
        -------
        log_likelihoods : dict of dict
            Keys will be model names and values will be dictionaries of
            log-liklihood values as returned by MNLLocationChoiceModel.fit.

        """
        with log_start_finish(
                'fit models in LCM group {}'.format(self.name), logger):
            return {
                name: self.models[name].fit(df, alternatives, current_choice)
                for name, df in self._iter_groups(choosers)}

    @property
    def fitted(self):
        """
        Whether all models in the group have been fitted.

        """
        return (all(m.fitted for m in self.models.values())
                if self.models else False)

    def predict(self, choosers, alternatives, debug=False):
        """
        Choose from among alternatives for a group of agents after
        segmenting the `choosers` table.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
            Only the first item in this table is used for determining
            agent probabilities of choosing alternatives.
            Must have a column matching the .segmentation_col attribute.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing.
        debug : bool
            If debug is set to true, will set the variable "sim_pdf" on
            the object to store the probabilities for mapping of the
            outcome.

        Returns
        -------
        choices : pandas.Series
            Mapping of chooser ID to alternative ID. Some choosers
            will map to a nan value when there are not enough alternatives
            for all the choosers.

        """
        logger.debug('start: predict models in LCM group {}'.format(self.name))
        results = []

        for name, df in self._iter_groups(choosers):
            choices = self.models[name].predict(df, alternatives, debug=debug)
            # remove chosen alternatives
            alternatives = alternatives.loc[~alternatives.index.isin(choices)]
            results.append(choices)

        logger.debug(
            'finish: predict models in LCM group {}'.format(self.name))
        return pd.concat(results) if results else pd.Series()

    def choosers_columns_used(self):
        """
        Columns from the choosers table that are used for filtering.

        """
        return list(toolz.unique(toolz.concat(
            m.choosers_columns_used() for m in self.models.values())))

    def alts_columns_used(self):
        """
        Columns from the alternatives table that are used for filtering.

        """
        return list(toolz.unique(toolz.concat(
            m.alts_columns_used() for m in self.models.values())))

    def interaction_columns_used(self):
        """
        Columns from the interaction dataset used for filtering and in
        the model. These may come originally from either the choosers or
        alternatives tables.

        """
        return list(toolz.unique(toolz.concat(
            m.interaction_columns_used() for m in self.models.values())))

    def columns_used(self):
        """
        Columns from any table used in the model. May come from either
        the choosers or alternatives tables.

        """
        return list(toolz.unique(toolz.concat(
            m.columns_used() for m in self.models.values())))


class SegmentedMNLLocationChoiceModel(object):
    """
    An MNL LCM group that allows segments to have different model expressions
    but otherwise share configurations.

    Parameters
    ----------
    segmentation_col
        Name of column in the choosers table that will be used for groupby.
    sample_size : int
        Number of choices to sample for estimating the model.
    choosers_fit_filters : list of str, optional
        Filters applied to choosers table before fitting the model.
    choosers_predict_filters : list of str, optional
        Filters applied to the choosers table before calculating
        new data points.
    alts_fit_filters : list of str, optional
        Filters applied to the alternatives table before fitting the model.
    alts_predict_filters : list of str, optional
        Filters applied to the alternatives table before calculating
        new data points.
    interaction_predict_filters : list of str, optional
        Filters applied to the merged choosers/alternatives table
        before predicting agent choices.
    estimation_sample_size : int, optional, whether to sample choosers
        during estimation (needs to be applied after choosers_fit_filters)
    choice_column : optional
        Name of the column in the `alternatives` table that choosers
        should choose. e.g. the 'building_id' column. If not provided
        the alternatives index is used.
    default_model_expr : str, iterable, or dict, optional
        A patsy model expression. Should contain only a right-hand side.
    name : str, optional
        An optional string used to identify the model in places.

    """
    def __init__(self, segmentation_col, sample_size,
                 choosers_fit_filters=None, choosers_predict_filters=None,
                 alts_fit_filters=None, alts_predict_filters=None,
                 interaction_predict_filters=None,
                 estimation_sample_size=None,
                 choice_column=None, default_model_expr=None, name=None):
        self.segmentation_col = segmentation_col
        self.sample_size = sample_size
        self.choosers_fit_filters = choosers_fit_filters
        self.choosers_predict_filters = choosers_predict_filters
        self.alts_fit_filters = alts_fit_filters
        self.alts_predict_filters = alts_predict_filters
        self.interaction_predict_filters = interaction_predict_filters
        self.estimation_sample_size = estimation_sample_size
        self.choice_column = choice_column
        self.default_model_expr = default_model_expr
        self._group = MNLLocationChoiceModelGroup(segmentation_col)
        self.name = (name if name is not None else
                     'SegmentedMNLLocationChoiceModel')

    @classmethod
    def from_yaml(cls, yaml_str=None, str_or_buffer=None):
        """
        Create a SegmentedMNLLocationChoiceModel instance from a saved YAML
        configuration. Arguments are mutally exclusive.

        Parameters
        ----------
        yaml_str : str, optional
            A YAML string from which to load model.
        str_or_buffer : str or file like, optional
            File name or buffer from which to load YAML.

        Returns
        -------
        SegmentedMNLLocationChoiceModel

        """
        cfg = yamlio.yaml_to_dict(yaml_str, str_or_buffer)

        default_model_expr = cfg['default_config']['model_expression']

        seg = cls(
            cfg['segmentation_col'],
            cfg['sample_size'],
            cfg['choosers_fit_filters'],
            cfg['choosers_predict_filters'],
            cfg['alts_fit_filters'],
            cfg['alts_predict_filters'],
            cfg['interaction_predict_filters'],
            cfg['estimation_sample_size'],
            cfg['choice_column'],
            default_model_expr,
            cfg['name'])

        if "models" not in cfg:
            cfg["models"] = {}

        for name, m in cfg['models'].items():
            m['model_expression'] = m.get(
                'model_expression', default_model_expr)
            m['sample_size'] = cfg['sample_size']
            m['choosers_fit_filters'] = None
            m['choosers_predict_filters'] = None
            m['alts_fit_filters'] = None
            m['alts_predict_filters'] = None
            m['interaction_predict_filters'] = \
                cfg['interaction_predict_filters']
            m['estimation_sample_size'] = cfg['estimation_sample_size']
            m['choice_column'] = cfg['choice_column']

            model = MNLLocationChoiceModel.from_yaml(
                yamlio.convert_to_yaml(m, None))
            seg._group.add_model(model)

        logger.debug(
            'loaded segmented LCM model {} from YAML'.format(seg.name))
        return seg

    def add_segment(self, name, model_expression=None):
        """
        Add a new segment with its own model expression.

        Parameters
        ----------
        name
            Segment name. Must match a segment in the groupby of the data.
        model_expression : str or dict, optional
            A patsy model expression that can be used with statsmodels.
            Should contain both the left- and right-hand sides.
            If not given the default model will be used, which must not be
            None.

        """
        logger.debug('adding LCM model {} to segmented model {}'.format(
            name, self.name))
        if not model_expression:
            if not self.default_model_expr:
                raise ValueError(
                    'No default model available, '
                    'you must supply a model expression.')
            model_expression = self.default_model_expr

        # we'll take care of some of the filtering this side before
        # segmentation
        self._group.add_model_from_params(
            name=name,
            model_expression=model_expression,
            sample_size=self.sample_size,
            choosers_fit_filters=None,
            choosers_predict_filters=None,
            alts_fit_filters=None,
            alts_predict_filters=None,
            interaction_predict_filters=self.interaction_predict_filters,
            estimation_sample_size=self.estimation_sample_size,
            choice_column=self.choice_column)

    def fit(self, choosers, alternatives, current_choice):
        """
        Fit and save models based on given data after segmenting
        the `choosers` table. Segments that have not already been explicitly
        added will be automatically added with default model.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
            Must have a column with the same name as the .segmentation_col
            attribute.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing,
            e.g. buildings.
        current_choice
            Name of column in `choosers` that indicates which alternative
            they have currently chosen.

        Returns
        -------
        log_likelihoods : dict of dict
            Keys will be model names and values will be dictionaries of
            log-liklihood values as returned by MNLLocationChoiceModel.fit.

        """
        logger.debug('start: fit models in segmented LCM {}'.format(self.name))
        choosers = util.apply_filter_query(choosers, self.choosers_fit_filters)

        unique = choosers[self.segmentation_col].unique()

        # Remove any existing segments that may no longer have counterparts
        # in the data. This can happen when loading a saved model and then
        # calling this method with data that no longer has segments that
        # were there the last time this was called.
        gone = set(self._group.models) - set(unique)
        for g in gone:
            del self._group.models[g]

        for x in unique:
            if x not in self._group.models:
                self.add_segment(x)

        alternatives = util.apply_filter_query(
            alternatives, self.alts_fit_filters)

        results = self._group.fit(choosers, alternatives, current_choice)
        logger.debug(
            'finish: fit models in segmented LCM {}'.format(self.name))
        return results

    @property
    def fitted(self):
        """
        Whether models for all segments have been fit.

        """
        return self._group.fitted

    def predict(self, choosers, alternatives, debug=False):
        """
        Choose from among alternatives for a group of agents after
        segmenting the `choosers` table.

        Parameters
        ----------
        choosers : pandas.DataFrame
            Table describing the agents making choices, e.g. households.
            Only the first item in this table is used for determining
            agent probabilities of choosing alternatives.
            Must have a column matching the .segmentation_col attribute.
        alternatives : pandas.DataFrame
            Table describing the things from which agents are choosing.
        debug : bool
            If debug is set to true, will set the variable "sim_pdf" on
            the object to store the probabilities for mapping of the
            outcome.

        Returns
        -------
        choices : pandas.Series
            Mapping of chooser ID to alternative ID. Some choosers
            will map to a nan value when there are not enough alternatives
            for all the choosers.

        """
        logger.debug(
            'start: predict models in segmented LCM {}'.format(self.name))
        choosers = util.apply_filter_query(
            choosers, self.choosers_predict_filters)
        alternatives = util.apply_filter_query(
            alternatives, self.alts_predict_filters)

        results = self._group.predict(choosers, alternatives, debug=debug)
        logger.debug(
            'finish: predict models in segmented LCM {}'.format(self.name))
        return results

    def _process_model_dict(self, d):
        """
        Remove redundant items from a model's configuration dict.

        Parameters
        ----------
        d : dict
            Modified in place.

        Returns
        -------
        dict
            Modified `d`.

        """
        del d['model_type']
        del d['sample_size']
        del d['choosers_fit_filters']
        del d['choosers_predict_filters']
        del d['alts_fit_filters']
        del d['alts_predict_filters']
        del d['interaction_predict_filters']
        del d['estimation_sample_size']
        del d['choice_column']

        if d['model_expression'] == self.default_model_expr:
            del d['model_expression']

        d["name"] = yamlio.to_scalar_safe(d["name"])

        return d

    def to_dict(self):
        """
        Returns a dict representation of this instance suitable for
        conversion to YAML.

        """
        return {
            'model_type': 'segmented_locationchoice',
            'name': self.name,
            'segmentation_col': self.segmentation_col,
            'sample_size': self.sample_size,
            'choosers_fit_filters': self.choosers_fit_filters,
            'choosers_predict_filters': self.choosers_predict_filters,
            'alts_fit_filters': self.alts_fit_filters,
            'alts_predict_filters': self.alts_predict_filters,
            'interaction_predict_filters': self.interaction_predict_filters,
            'estimation_sample_size': self.estimation_sample_size,
            'choice_column': self.choice_column,
            'default_config': {
                'model_expression': self.default_model_expr,
            },
            'fitted': self.fitted,
            'models': {
                yamlio.to_scalar_safe(name):
                    self._process_model_dict(m.to_dict())
                for name, m in self._group.models.items()
            }
        }

    def to_yaml(self, str_or_buffer=None):
        """
        Save a model respresentation to YAML.

        Parameters
        ----------
        str_or_buffer : str or file like, optional
            By default a YAML string is returned. If a string is
            given here the YAML will be written to that file.
            If an object with a ``.write`` method is given the
            YAML will be written to that object.

        Returns
        -------
        j : str
            YAML is string if `str_or_buffer` is not given.

        """
        logger.debug('serializing segmented LCM {} to YAML'.format(self.name))
        return yamlio.convert_to_yaml(self.to_dict(), str_or_buffer)

    def choosers_columns_used(self):
        """
        Columns from the choosers table that are used for filtering.

        """
        return list(toolz.unique(toolz.concatv(
            util.columns_in_filters(self.choosers_predict_filters),
            util.columns_in_filters(self.choosers_fit_filters))))

    def alts_columns_used(self):
        """
        Columns from the alternatives table that are used for filtering.

        """
        return list(toolz.unique(toolz.concatv(
            util.columns_in_filters(self.alts_predict_filters),
            util.columns_in_filters(self.alts_fit_filters))))

    def interaction_columns_used(self):
        """
        Columns from the interaction dataset used for filtering and in
        the model. These may come originally from either the choosers or
        alternatives tables.

        """
        return self._group.interaction_columns_used()

    def columns_used(self):
        """
        Columns from any table used in the model. May come from either
        the choosers or alternatives tables.

        """
        return list(toolz.unique(toolz.concatv(
            self.choosers_columns_used(),
            self.alts_columns_used(),
            self.interaction_columns_used(),
            util.columns_in_formula(self.default_model_expr),
            [self.segmentation_col])))

    @classmethod
    def fit_from_cfg(cls, choosers, chosen_fname, alternatives, cfgname):
        """
        Parameters
        ----------
        choosers : DataFrame
            A dataframe of rows of agents which have locations assigned.
        chosen_fname : string
            A string indicating the column in the choosers dataframe which
            gives which location the choosers have chosen.
        alternatives : DataFrame
            A dataframe of locations which should include the chosen locations
            from the choosers dataframe as well as some other locations from
            which to sample.  Values in choosers[chosen_fname] should index
            into the alternatives dataframe.
        cfgname : string
            The name of the yaml config file from which to read the location
            choice model.

        Returns
        -------
        lcm : SegmentedMNLLocationChoiceModel which was used to fit
        """
        logger.debug('start: fit from configuration {}'.format(cfgname))
        lcm = cls.from_yaml(str_or_buffer=cfgname)
        lcm.fit(choosers, alternatives, choosers[chosen_fname])
        for k, v in lcm._group.models.items():
            print("LCM RESULTS FOR SEGMENT %s\n" % str(k))
            v.report_fit()
        lcm.to_yaml(str_or_buffer=cfgname)
        logger.debug('finish: fit from configuration {}'.format(cfgname))
        return lcm

    @classmethod
    def predict_from_cfg(cls, movers, locations, cfgname,
                         location_ratio=2.0, debug=False):
        """
        Simulate the location choices for the specified choosers

        Parameters
        ----------
        movers : DataFrame
            A dataframe of agents doing the choosing.
        locations : DataFrame
            A dataframe of locations which the choosers are locating in and which
            have a supply.
        cfgname : string
            The name of the yaml config file from which to read the location
            choice model.
        location_ratio : float
            Above the location ratio (default of 2.0) of locations to choosers, the
            locations will be sampled to meet this ratio (for performance reasons).

        Returns
        -------
        choices : pandas.Series
            Mapping of chooser ID to alternative ID. Some choosers
            will map to a nan value when there are not enough alternatives
            for all the choosers.
        lcm : SegmentedMNLLocationChoiceModel which was used to predict
        """
        logger.debug('start: predict from configuration {}'.format(cfgname))
        lcm = cls.from_yaml(str_or_buffer=cfgname)

        if len(locations) > len(movers) * location_ratio:
            logger.info("Location ratio exceeded: %d locations and only %d choosers" %
                        (len(locations), len(movers)))
            idxes = np.random.choice(locations.index, size=len(movers) * location_ratio,
                                     replace=False)
            locations = locations.loc[idxes]
            logger.info("  after sampling %d locations are available\n" % len(locations))

        new_units = lcm.predict(movers, locations, debug=debug)
        print("Assigned %d choosers to new units" % len(new_units.dropna()))
        logger.debug('finish: predict from configuration {}'.format(cfgname))
        return new_units, lcm



"""
This module exists to help us with some performance headaches in patsy.
Specifically with categorical data. A lot of urbansim models have
boolean columns via expressions like I(year_built < 1940). patsy's
categorical code checks every single value in those columns to see
whether it is nan, but if we know the data are pandas.Series instances
we should be able to check the dtype to see if it is boolean and,
if so, take appropriate specific action that avoids looking at every
value.

The function ``patch_patsy`` monkeypatches a function and a method
in patsy with replacements that check the input data's dtype for
boolean and act appropriately.

"""
import numpy as np
import pandas as pd
from patsy import PatsyError
from patsy.categorical import _CategoricalBox, C
from patsy.util import (SortAnythingKey,
                        safe_scalar_isnan,
                        iterable)


# replaces function patsy.categorical.categorical_to_int
def categorical_to_int(data, levels, NA_action, origin=None):
    assert isinstance(levels, tuple)
    # In this function, missing values are always mapped to -1
    if isinstance(data, pd.Categorical):
        data_levels_tuple = tuple(data.levels)
        if not data_levels_tuple == levels:
            raise PatsyError("mismatching levels: expected %r, got %r"
                             % (levels, data_levels_tuple), origin)
        # pd.Categorical also uses -1 to indicate NA, and we don't try to
        # second-guess its NA detection, so we can just pass it back.
        return data.labels
    elif hasattr(data, 'dtype') and hasattr(data, 'astype') and \
            np.issubdtype(data.dtype, np.bool_):
        return data.astype('int')
    if isinstance(data, _CategoricalBox):
        if data.levels is not None and tuple(data.levels) != levels:
            raise PatsyError("mismatching levels: expected %r, got %r"
                             % (levels, tuple(data.levels)), origin)
        data = data.data
    if hasattr(data, "shape") and len(data.shape) > 1:
        raise PatsyError("categorical data must be 1-dimensional",
                         origin)
    if not iterable(data) or isinstance(data, basestring):
        raise PatsyError("categorical data must be an iterable container")
    try:
        level_to_int = dict(zip(levels, xrange(len(levels))))
    except TypeError:
        raise PatsyError("Error interpreting categorical data: "
                         "all items must be hashable", origin)
    out = np.empty(len(data), dtype=int)
    for i, value in enumerate(data):
        if NA_action.is_categorical_NA(value):
            out[i] = -1
        else:
            try:
                out[i] = level_to_int[value]
            except KeyError:
                SHOW_LEVELS = 4
                level_strs = []
                if len(levels) <= SHOW_LEVELS:
                    level_strs += [repr(level) for level in levels]
                else:
                    level_strs += [repr(level)
                                   for level in levels[:SHOW_LEVELS//2]]
                    level_strs.append("...")
                    level_strs += [repr(level)
                                   for level in levels[-SHOW_LEVELS//2:]]
                level_str = "[%s]" % (", ".join(level_strs))
                raise PatsyError("Error converting data to categorical: "
                                 "observation with value %r does not match "
                                 "any of the expected levels (expected: %s)"
                                 % (value, level_str), origin)
            except TypeError:
                raise PatsyError("Error converting data to categorical: "
                                 "encountered unhashable value %r"
                                 % (value,), origin)
    if isinstance(data, pd.Series):
        out = pd.Series(out, index=data.index)
    return out


# replaces method patsy.categorical.CategoricalSniffer.sniff
def sniff(self, data):
    if hasattr(data, "contrast"):
        self._contrast = data.contrast
    # returns a bool: are we confident that we found all the levels?
    if isinstance(data, pd.Categorical):
        # pandas.Categorical has its own NA detection, so don't try to
        # second-guess it.
        self._levels = tuple(data.levels)
        return True
    elif hasattr(data, 'dtype') and np.issubdtype(data.dtype, np.bool_):
        self._level_set = set([True, False])
        return True
    if isinstance(data, _CategoricalBox):
        if data.levels is not None:
            self._levels = tuple(data.levels)
            return True
        else:
            # unbox and fall through
            data = data.data
    for value in data:
        if self._NA_action.is_categorical_NA(value):
            continue
        if value is True or value is False:
            self._level_set.update([True, False])
        else:
            try:
                self._level_set.add(value)
            except TypeError:
                raise PatsyError("Error interpreting categorical data: "
                                 "all items must be hashable",
                                 self._origin)
    # If everything we've seen is boolean, assume that everything else
    # would be too. Otherwise we need to keep looking.
    return self._level_set == set([True, False])


def patch_patsy():
    """
    Replace the function patsy.categorical.categorical_to_int
    (in multiple places) and the method
    patsy.categorical.CategoricalSniffer.sniff with replacements
    of our own that have special handling for arrays with
    boolean dtypes.

    """
    import patsy
    patsy.categorical.categorical_to_int = categorical_to_int
    patsy.build.categorical_to_int = categorical_to_int
    patsy.categorical.CategoricalSniffer.sniff = sniff
